"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    }
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
/** @module stately-react */
var React = require("react");
var stately_async_1 = require("stately-async");
var Controllable_1 = require("./Controllable");
var _a = Controllable_1.createControllableContext(stately_async_1.statelyAsyncReducer, stately_async_1.statelyAsyncMiddleware), AsyncController = _a.Controller, AsyncControllable = _a.Controllable;
exports.AsyncController = AsyncController;
/**
 * Props: {@link AsyncProps}
 *
 * Context Providers: [Controller]{@link createControllableContext}<{@link AsyncState}>
 *
 * A render-prop component that injects an {@link AsyncLifecycle} for any {@link AsyncOperation} into the component tree.
 * It executes the given `AsyncOperation` on mount, and again whenever the given `params` change.
 * The {@link AsyncState} representing the operation is passed to this component's `children` as the sole parameter.
 * It should be used whenever a component needs asynchronously-loaded data to render, such as search results or an entity from a REST service.
 *
 * Each instance of `Async` will own a unique `AsyncLifecycle`.
 * This way, multiple instances using the same `AsyncOperation` will not conflict.
 * The component cleans up after itself by destroying the `AsyncLifecycle` on unmount.
 *
 * The following example wraps a `<SearchResults>` component, handling the execution and state management of the asynchronous search call:
 * ```
 * // type doSearch = (p1: number, p2: string) => Promise<SearchResults>
 *
 * <Async operation={doSearch} params={[123, 'abc']}>
 *   {state =>
 *     <div>
 *       {
 *         // If the Promise is rejected, render the error.
 *         state.error ? <ErrorMessage error={state.error} />
 *           // If the Promise resolves, render the SearchResults.
 *           : state.data ? <SearchResults results={state.data} />
 *           // If the operation is active, render a loading indicator.
 *           : state.status === 'active' && <LoadingSpinner />
 *       }
 *     </div>}
 * </Async>
 * ```
 */
var Async = /** @class */ (function (_super) {
    __extends(Async, _super);
    function Async(props) {
        var _this = _super.call(this, props) || this;
        _this.asyncLifecycle = stately_async_1.asyncLifecycle(props.operation);
        return _this;
    }
    Async.prototype.render = function () {
        var _this = this;
        var _a = this.asyncLifecycle, selector = _a.selector, call = _a.call, destroy = _a.destroy;
        var params = this.props.params;
        return (React.createElement(AsyncControllable, null, function (state, dispatch) { return (React.createElement(LifecycleAsync, { params: params, state: selector(state), call: function () {
                var params = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    params[_i] = arguments[_i];
                }
                dispatch(call.apply(void 0, params));
            }, destroy: function () {
                dispatch(destroy());
            } }, _this.props.children)); }));
    };
    return Async;
}(React.Component));
exports.Async = Async;
var paramsChanged = function (prev, next) {
    return prev !== next &&
        (!prev ||
            !next ||
            prev.length !== next.length ||
            !prev.every(function (value, index) { return value === next[index]; }));
};
var LifecycleAsync = /** @class */ (function (_super) {
    __extends(LifecycleAsync, _super);
    function LifecycleAsync() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    LifecycleAsync.prototype.render = function () {
        var _a = this.props, children = _a.children, state = _a.state;
        return children(state);
    };
    LifecycleAsync.prototype.componentDidMount = function () {
        this.maybeCall();
    };
    LifecycleAsync.prototype.componentDidUpdate = function () {
        this.maybeCall();
    };
    LifecycleAsync.prototype.maybeCall = function () {
        var _a;
        // people are likely to pass params as an array literal, which is always a new reference
        if (paramsChanged(this.params, this.props.params)) {
            this.params = this.props.params;
            if (this.params) {
                (_a = this.props).call.apply(_a, this.params);
            }
        }
    };
    LifecycleAsync.prototype.componentWillUnmount = function () {
        this.props.destroy();
    };
    return LifecycleAsync;
}(React.Component));
//# sourceMappingURL=Async.js.map